@@ -5,7 +5,7 @@
 ;; Author: Yuan Fu <casouri@gmail.com>
 ;; Maintainer: Yuan Fu <casouri@gmail.com>
 ;; URL: https://sr.ht/~casouri/xeft
-;; Version: 3.3
+;; Version: 3.4
 ;; Keywords: Applications, Note, Searching
 ;; Package-Requires: ((emacs "26.0"))
 
@@ -40,15 +40,18 @@
 ;;
 ;; Note that:
 ;;
-;; 1. Xeft only looks for first-level files in ‘xeft-directory’. Files
+;; 1. Xeft ignores search phrases shorter than three characters,
+;;    unless they are CJK characters.
+;;
+;; 2. Xeft only looks for first-level files in ‘xeft-directory’. Files
 ;;    in sub-directories are not searched unless ‘xeft-recursive’ is
 ;;    non-nil.
 ;;
-;; 2. Xeft creates a new file by using the search phrase as the
+;; 3. Xeft creates a new file by using the search phrase as the
 ;;    filename and title. If you want otherwise, redefine
 ;;    ‘xeft-create-note’ or ‘xeft-filename-fn’.
 ;;
-;; 3. Xeft saves the current window configuration before switching to
+;; 4. Xeft saves the current window configuration before switching to
 ;;    Xeft buffer. When Xeft buffer is killed, Xeft restores the saved
 ;;    window configuration.
 ;;
@@ -171,7 +174,11 @@ indexed in the database, simply delete the database and start
 xeft again.
 
 If this is not flexible enough, take a look at
-‘xeft-file-filter’."
+‘xeft-file-filter’.
+
+Changing this variable along doesn’t remove already-indexed files
+from the database, you need to delete the database on disk and
+let xeft recreate it."
   :type '(list string))
 
 (defcustom xeft-file-filter #'xeft-default-file-filter
@@ -180,7 +187,11 @@ If this is not flexible enough, take a look at
 If ‘xeft-ignore-extension’ is not flexible enough, customize this
 function to filter out unwanted files. This function should take
 the absolute path of a file and return t/nil indicating
-keeping/excluding the file from indexing."
+keeping/excluding the file from indexing.
+
+Changing this variable along doesn’t remove already-indexed files
+from the database, you need to delete the database on disk and
+let xeft recreate it."
   :type 'function)
 
 (defcustom xeft-directory-filter #'xeft-default-directory-filter
@@ -188,7 +199,11 @@ keeping/excluding the file from indexing."
 
 This function is useful when ‘xeft-recursive’ is non-nil, and you
 want to exclude certain directories (and its enclosing files)
-from indexing."
+from indexing.
+
+Changing this variable along doesn’t remove already-indexed files
+from the database, you need to delete the database on disk and
+let xeft recreate it."
   :type 'function)
 
 (defcustom xeft-title-function #'xeft-default-title
@@ -203,18 +218,35 @@ point at the beginning of body text (ie, end of title)."
   :type 'function)
 
 (defcustom xeft-recursive nil
-  "If non-nil, xeft searches for file recursively.
+  "If non-nil, xeft searches for files recursively.
 
 Xeft doesn’t follow symlinks and ignores inaccessible
 directories. Customize ‘xeft-directory-filter’ to exclude
-subdirectories."
-  :type 'boolean)
+subdirectories. Set this variable to ‘follow-symlinks’ to follow
+symlinks, note that this might lead to infinite recursion.
+
+Changing this variable along doesn’t remove already-indexed files
+from the database, you need to delete the database on disk and
+let xeft recreate it."
+  :type '(choice (const :tag "No" nil)
+                 (const :tag "Yes" t)
+                 (const :tag "Yes, and follow symlinks"
+                        follow-symlinks )))
 
 (defcustom xeft-file-list-function #'xeft--file-list
   "A function that returns files that xeft should search from.
-This function takes no arguments and return a list of absolute paths."
+This function takes no arguments and return a list of absolute paths.
+
+Changing this variable along doesn’t remove already-indexed files
+from the database, you need to delete the database on disk and
+let xeft recreate it."
   :type 'function)
 
+(defcustom xeft-compile-args nil
+  "Compilation flags for compiling xeft-lite.
+If non-nil, Xeft passes this list of args to make to compile xeft-lite."
+  :type '(list string))
+
 ;;; Compile
 
 (defun xeft--compile-module ()
@@ -225,11 +257,13 @@ This function takes no arguments and return a list of absolute paths."
   (let* ((default-directory
           (file-name-directory
            (locate-library "xeft.el" t)))
-         (prefix (concat "PREFIX="
-                         (read-string "PREFIX (empty by default): ")))
+         (args
+          (or xeft-compile-args
+              (list (concat "PREFIX="
+                            (read-string "PREFIX (empty by default): ")))))
          (buffer (get-buffer-create "*xeft compile*")))
     (if (zerop (let ((inhibit-read-only t))
-                 (call-process "make" nil buffer t prefix)))
+                 (apply #'call-process "make" nil buffer t args)))
         (progn (message "Successfully compiled the module :-D") t)
       (pop-to-buffer buffer)
       (compilation-mode)
@@ -347,6 +381,25 @@ If success return non-nil, otherwise return nil."
     (insert "\n\nInsert search phrase and press RET to search.")
     (goto-char (point-min))))
 
+(defvar xeft-periodic-commit-timer-idle-delay 60
+  "Whenever Emacs is idle for this many seconds, commit changes to DB.
+
+This variable must be customized before the first call to ‘xeft’, in
+which we setup the idle timer.")
+
+(defvar xeft--periodic-commit-timer nil
+  "Timer for the periodic commit routine (‘xeft--periodic-commit’).")
+
+(defun xeft--periodic-commit ()
+  "Commit DB modifications to disk by closing the DB.
+
+This is called periodically so modifications made to DB are saved even
+if there’s a Emacs crash."
+  (when (fboundp 'xapian-lite-close-database)
+    (xapian-lite-close-database xeft-database))
+  ;; Immediately query something so we re-open the DB to prevent
+  ;; someone else from locking it for themselves. __MY__ database!
+  (xapian-lite-query-term "AAAAAAAAAAA" xeft-database 0 10))
 
 ;;; Userland
 
@@ -374,7 +427,12 @@ If success return non-nil, otherwise return nil."
     ;; some file are changed without Xeft noticing.
     (xeft-full-reindex)
     ;; Also regenerate newest file cache, for the same reason as above.
-    (xeft--front-page-cache-refresh)))
+    (xeft--front-page-cache-refresh)
+    (unless xeft--periodic-commit-timer
+      (setq xeft--periodic-commit-timer
+            (run-with-idle-timer
+             xeft-periodic-commit-timer-idle-delay
+             t #'xeft--periodic-commit)))))
 
 (defun xeft-create-note ()
   "Create a new note with the current search phrase as the title."
@@ -428,11 +486,18 @@ If success return non-nil, otherwise return nil."
     (end-of-line)))
 
 (defun xeft-full-reindex ()
-  "Do a full reindex of all files."
+  "Do a full reindex of all files.
+
+This function only reindex files but doesn’t delete the database
+and recreate it. So if you changed the filter
+functions (‘xeft-file-filter’, etc), and want to remove the
+now-excluded files from the database, you need to manually delete
+the database."
   (interactive)
   (condition-case _
       (dolist (file (funcall xeft-file-list-function))
         (xapian-lite-reindex-file file xeft-database))
+
     (xapian-lite-database-lock-error
      (message "The Xeft database is locked (maybe there is another Xeft instance running) so we will skip indexing for now"))
     (xapian-lite-database-corrupt-error
@@ -467,7 +532,10 @@ If SELECT is non-nil, select the buffer after displaying it."
     (if (and (window-live-p xeft--preview-window)
              (not (eq xeft--preview-window (selected-window))))
         (with-selected-window xeft--preview-window
-          (switch-to-buffer buffer))
+          (switch-to-buffer buffer)
+          (when keyword-list
+            (let ((case-fold-search t))
+              (search-forward (car keyword-list) nil t))))
       (setq xeft--preview-window
             (display-buffer
              buffer '((display-buffer-use-some-window
@@ -561,7 +629,7 @@ title."
   (let ((bol (point)) title)
     (end-of-line)
     (setq title (buffer-substring-no-properties bol (point)))
-    (if (eq title "")
+    (if (equal title "")
         (file-name-base file)
       title)))
 
@@ -671,6 +739,15 @@ The score is the number of search phrases that appears in TITLE."
   (interactive)
   (xeft-refresh t))
 
+(defun xeft--file-name-extension (path)
+  "Return the extension part of PATH.
+This differs from ‘file-name-extension’ in that it doesn’t remove
+trailing \"version strings\"."
+  (let ((filename (file-name-nondirectory path)))
+    (when (string-match (rx (not ".") "." (group (* (not "."))) eos)
+                        filename)
+      (match-string 1 filename))))
+
 (defun xeft-default-file-filter (file)
   "Return nil if FILE should be ignored.
 
@@ -680,7 +757,7 @@ directories, dot files, and files matched by
   (and (file-regular-p file)
        (not (string-prefix-p
              "." (file-name-base file)))
-       (not (member (file-name-extension file)
+       (not (member (xeft--file-name-extension file)
                     xeft-ignore-extension))))
 
 (defun xeft-default-directory-filter (dir)
@@ -697,7 +774,8 @@ files and directories and check for ‘xeft-ignore-extension’."
    xeft-file-filter
    (if xeft-recursive
        (directory-files-recursively
-        xeft-directory "" nil xeft-directory-filter)
+        xeft-directory "" nil xeft-directory-filter
+        (eq xeft-recursive 'follow-symlinks))
      (directory-files
       xeft-directory t nil t))))
 
@@ -789,6 +867,13 @@ non-nil, display all results."
                  ;; 16 is just larger than 15, so we will know it when
                  ;; there are more results.
                  0 (if full 2147483647 16))))
+        ;; Sort results by last access time.
+        (setq file-list
+              (cl-stable-sort file-list
+                              (lambda (f1 f2)
+                                (not (time-less-p
+                                      (nth 4 (file-attributes f1))
+                                      (nth 4 (file-attributes f2)))))))
         (when (and (null full) (> (length file-list) 15))
           (setq file-list (cl-subseq file-list 0 15)
                 list-clipped t))
